/**
 * Global function to report errors.
 */

/**
 * Adds context to the given scope. Note that `context.culprit` is a mixmax specific property; if present, it will
 * be set as a tag. If `context.fingerprint` is not present, `context.culprit` will also be added to the fingerprint
 * for the event to disambiguate errors that would otherwise be grouped together due to their stacktrace.
 *
 * See {@link https://docs.sentry.io/platforms/javascript/enriching-events/scopes/#local-scopes}.
 *
 * @param scope
 * @param context
 */
function configureScope(scope, { user, tags, extra, fingerprint, culprit } = {}) {
  if (user) scope.setUser(user);
  if (tags) scope.setTags(tags);
  if (extra) scope.setExtras(extra);
  if (fingerprint) scope.setFingerprint(fingerprint);
  if (culprit) {
    scope.setTag('culprit', culprit);
    if (!fingerprint) {
      // See https://docs.sentry.io/platforms/node/data-management/event-grouping/sdk-fingerprinting/
      scope.setFingerprint(['{{ default }}', culprit]);
    }
  }
}

/**
 * Wrapper for Sentry's captureMessage.
 */
const captureMessage = function(message, options) {
  // Log to console locally to help debugging.
  console.error(message, options);

  Sentry.withScope((scope) => {
    configureScope(scope, options);
    Sentry.captureMessage(message);
  });
};
/**
 * Wrapper for Sentry's captureException.
 */
const captureException = function(error, options) {
  // Log to console locally to help debugging.
  console.error(error.stack || error, options);

  Sentry.withScope((scope) => {
    configureScope(scope, options);
    Sentry.captureException(error);
  });
};

window.error = {
  captureMessage,
  captureException,
};
