
const ExtensionMessageBus = (function() {
  const listeners = {};

  const tabQueryRedirects = {};

  function redirectKeyForQuery(tabQuery) {
    return JSON.stringify(tabQuery);
  }

  const pendingSends = {};

  function trigger(request, sender) {
    $(listeners).trigger(request.method, {
      invocationId: request.invocationId,
      payload: request.payload,
      sender: {
        id: sender.tab.id,
        url: sender.tab.url,
        canonicalUrl: request.canonicalUrl,
      },
    });
  }

  function sendToTabs(method, payload, tabs, tabQuery, options) {
    if (tabs && tabs.length) {
      tabs.forEach(function(tab) {
        chrome.tabs.sendMessage(tab.id, {
          method,
          payload,
        });
      });

      if (tabs.length === 1 && options.activateAfterSending) {
        chrome.tabs.update(tabs[0].id, {
          active: true,
        });
      }
    } else if (tabQuery.canonicalUrl || tabQuery.url) {
      const redirectKey = redirectKeyForQuery(tabQuery);

      let redirectedQuery = tabQueryRedirects[redirectKey];
      if (redirectedQuery) {
        const pendingSendsForTab = pendingSends[redirectedQuery.id];
        if (pendingSendsForTab) {
          pendingSendsForTab.push([method, payload, redirectedQuery]);
        } else {
          send(method, payload, redirectedQuery, options);
        }
        return;
      }

      chrome.tabs.create(
        {
          url: tabQuery.canonicalUrl || tabQuery.url,
        },
        function(tab) {
          redirectedQuery = {
            id: tab.id,
            url: tabQuery.url,
            canonicalUrl: tabQuery.canonicalUrl,
          };
          tabQueryRedirects[redirectKey] = redirectedQuery;
          const pendingSendsForTab = (pendingSends[redirectedQuery.id] = []);
          pendingSendsForTab.push([method, payload, redirectedQuery]);
        }
      );
    }
  }

  function send(method, payload, tabQuery , options ) {
    tabQuery = tabQuery || {}; 
    options = _.defaults({}, options, { activateAfterSending: false });

    let redirectedQuery = tabQueryRedirects[redirectKeyForQuery(tabQuery)];
    if (redirectedQuery) {
      while (redirectedQuery) {
        tabQuery = redirectedQuery;
        redirectedQuery = tabQueryRedirects[redirectKeyForQuery(tabQuery)];
      }
      const pendingSendsForTab = pendingSends[tabQuery.id];
      if (pendingSendsForTab) {
        pendingSendsForTab.push([method, payload, tabQuery]);
        return;
      }
    }

    if (tabQuery.id) {
      chrome.tabs.get(tabQuery.id, function(tab) {
        if (chrome.runtime.lastError) {
        }
        sendToTabs(method, payload, tab && [tab], tabQuery, options);
      });
    } else {
      chrome.tabs.query({ url: tabQuery.url, active: tabQuery.active }, function(tabs) {
        if (!_.isUndefined(tabQuery.notId)) {
          tabs = _.reject(tabs, function(tab) {
            return tab.id === tabQuery.notId;
          });
        }
        sendToTabs(method, payload, tabs, tabQuery, options);
      });
    }
  }

  function on(method, callback) {
    $(listeners).on(method, function(evt, context) {
      if (context.invocationId) {
        callback(context.payload, context.sender, function(response) {
          send('_IN_REPLY_TO_' + context.invocationId, response);
        });
        delete context.invocationId; 
      } else {
        callback(context.payload, context.sender);
      }
    });
  }

  on('listenerReady', function(method, sender) {
    let pendingSendsForTab = pendingSends[sender.id];
    if (pendingSendsForTab) {
      const completedSends = [];
      pendingSendsForTab.forEach(function(sendArguments) {
        if (sendArguments[0] === method) {
          send.apply(null, sendArguments);
          completedSends.push(sendArguments);
        }
      });
      pendingSendsForTab = pendingSendsForTab.filter(function(sendArguments) {
        return completedSends.indexOf(sendArguments) === -1;
      });
      if (pendingSendsForTab.length) {
        pendingSends[sender.id] = pendingSendsForTab;
      } else {
        delete pendingSends[sender.id];
      }
    }

  });

  return {
    send,

    on,

    trigger,
  };
})();


function bundleError(err) {
  const errObj = typeof err === 'string' ? new Error(err) : err;

  return {
    message: errObj.message,
    name: errObj.name,
    props: Object.assign(
      {
        stack: errObj.stack,
      },
      errObj
    ),
  };
}

const XHR_ERROR_PROPS = [
  'readyState',
  'responseText',
  'responseType',
  'responseURL',
  'status',
  'statusText',
];

class XHRError extends Error {
  constructor(message, xhr) {
    super(message);
    for (const prop of XHR_ERROR_PROPS) {
      this[prop] = xhr[prop];
    }
  }
}

async function getBlobFromObjectURL(objectURL) {
  const xhr = new XMLHttpRequest();
  xhr.open('GET', objectURL, true);
  xhr.responseType = 'blob';

  return new Promise((resolve, reject) => {
    xhr.addEventListener('load', () => {
      if (xhr.status === 200) {
        resolve(xhr.response);
      } else {
        reject(new XHRError('Blob resolution got non-200 response code.', xhr));
      }
    });

    xhr.addEventListener('error', () => {
      reject(new XHRError('Blob resolution failed.', xhr));
    });

    xhr.send();
  });
}

const BLOB_PROTOCOL_PATTERN = /blob/i;

function isBlobURL(url) {
  return !!url && BLOB_PROTOCOL_PATTERN.test(new URL(url).protocol);
}

const ResponseCache = (function() {
  const blobResponseCache = new Map();
  const RESPONSE_CACHE_TTL = 5 * 60 * 1000;

  function cacheBlobResponse(url, blob) {
    const blobURL = URL.createObjectURL(blob);
    blobResponseCache.set(url, { time: Date.now(), blobURL });
    return blobURL;
  }

  async function getCachedBlobResponse(url) {
    const response = blobResponseCache.get(url);
    if (!response) return null;
    return fetch(response.blobURL).then((resp) => resp.blob());
  }

  function removeCachedBlobResponse(url) {
    const response = blobResponseCache.get(url);
    if (response && response.blobURL) URL.revokeObjectURL(response.blobURL);
    blobResponseCache.delete(url);
  }

  setInterval(() => {
    const threshold = Date.now() - RESPONSE_CACHE_TTL;
    for (const [url, response] of blobResponseCache) {
      if (response.time < threshold) {
        removeCachedBlobResponse(url);
      }
    }
  }, RESPONSE_CACHE_TTL);

  return {
    cacheBlobResponse,
    getCachedBlobResponse,
    removeCachedBlobResponse,
  };
})();







function wrapAsyncMessageHandler(handler, willSendResponse) {
  return function(request, messageSender, sendMessage) {
    handler(request, messageSender, sendMessage);
    return willSendResponse;
  };
}


const UploaderProxy = (function() {







  class FileNotFoundError extends Error {
    constructor(fileRef, message = 'Referenced file not found') {
      super(message);
      this.file = fileRef;
    }
  }

  const CLOUDINARY_UPLOAD_URL = 'https://api.cloudinary.com/v1_1/mixmax/upload';

  const throttledCredentialsGet = _.throttle(
    async ({ credentials, region }) => {
      AWS.config.update({ region });
      return credentials.getPromise();
    },
    500,
    { trailing: false }
  ); 

  async function uploadFileToS3({ bucketParams, uploadParams, uploadOptions, refetchCredentials }) {
    const bucket = await loadAttachmentsBucket(bucketParams, refetchCredentials);
    const upload = bucket.upload(uploadParams, uploadOptions);
    return upload.promise();
  }

  async function loadAttachmentsBucket(bucketParams, refetchCredentials) {
    const hydratedBucketParams = {
      ..._.omit(bucketParams, 'credentialsParams'),
      credentials: new AWS.CognitoIdentityCredentials(bucketParams.credentialsParams),
    };
    if (refetchCredentials) {
      await throttledCredentialsGet(hydratedBucketParams);
    }
    return new AWS.S3(hydratedBucketParams);
  }

  async function getFile(fileRef) {
    if (!fileRef.url && !fileRef.objectURL) {
      throw new FileNotFoundError(fileRef);
    }

    let blob;
    if (fileRef.url) {
      blob = await ResponseCache.getCachedBlobResponse(fileRef.url);
    }

    if (!blob && fileRef.objectURL) {
      blob = await getBlobFromObjectURL(fileRef.objectURL);
    }

    if (!blob) {
      throw new FileNotFoundError(fileRef);
    } else if (!blob.size) {
      throw new FileNotFoundError(fileRef, 'Fetched empty blob');
    }
    return new File([blob], fileRef.name);
  }

  async function handleUploadFileToS3(request, messageSender, sendResponse) {
    try {
      const { config } = request;
      const file = await getFile(config.fileRef);
      if (!file) {
        return void sendResponse({
          success: false,
          payload: bundleError(new FileNotFoundError(config.fileRef)),
        });
      }

      config.uploadParams.Body = file;
      const result = await uploadFileToS3(config);
      sendResponse({ success: true, payload: result });
      ResponseCache.removeCachedBlobResponse(config.fileRef.url);
    } catch (err) {
      sendResponse({ success: false, payload: bundleError(err) });
    }
  }

  function uploadToCloudinary({ file, folder, uploadPreset }, sendUpdate) {
    const fd = new FormData();
    fd.append('upload_preset', uploadPreset);
    fd.append('folder', folder);
    fd.append('file', file);

    const xhr = new XMLHttpRequest();
    xhr.open('POST', CLOUDINARY_UPLOAD_URL, true);
    xhr.setRequestHeader('X-Requested-With', 'XMLHttpRequest');
    xhr.upload.addEventListener('progress', (e) => {
      const percentage = e.lengthComputable ? Math.round((e.loaded * 100.0) / e.total) : 0;
      sendUpdate({ percentage });
    });

    xhr.addEventListener('load', () => {
      const response = JSON.parse(xhr.responseText);
      sendUpdate({ success: true, payload: response });
    });

    xhr.addEventListener('error', () => {
      sendUpdate({ success: false, payload: bundleError('Error uploading to cloudinary') });
    });

    xhr.send(fd);
  }

  async function handleUploadToCloudinary(config, sendUpdate) {
    try {
      const file = await getFile(config.fileRef);
      if (!file) {
        return void sendUpdate({
          success: false,
          payload: bundleError(new FileNotFoundError(config.fileRef)),
        });
      }

      uploadToCloudinary(
        {
          file,
          folder: config.folder,
          uploadPreset: config.uploadPreset,
        },
        sendUpdate
      );
    } catch (err) {
      sendUpdate({ success: false, payload: bundleError(err) });
    }
  }

  function handlePortConnect(port) {
    port.onMessage.addListener(({ config, method, id }) => {
      const sendUpdate = ({ success, percentage, payload }) => {
        port.postMessage({ id, success, percentage, payload });
      };

      if (method !== 'uploadToCloudinary') {
        sendUpdate({
          success: false,
          payload: bundleError(new TypeError(`invalid method ${method}`)),
        });
        return;
      }

      handleUploadToCloudinary(config, sendUpdate);
    });
  }

  return {
    handleUploadFileToS3: wrapAsyncMessageHandler(handleUploadFileToS3, true),
    handlePortConnect,
  };
})();


chrome.browserAction.onClicked.addListener(function(tab) {
  chrome.windows.get(tab.windowId, function(win) {
    chrome.storage.local.get(null , function(storage) {
      openComposeWindow(win, tab, storage);
    });
  });
});

function openComposeWindow(win, tab, storage, params) {
  const onboardedAccounts = getOnboardedAccounts(storage);
  if (onboardedAccounts.length === 0) {
    return void refreshBrowser('gmail', true );
  }

  const userEmail = _getUserEmail(storage);

  const messageId = generateMessageId();
  let url =
    Environment.getComposeUrl() +
    '/message/' +
    messageId +
    '?user=' +
    encodeURIComponent(userEmail);

  if (onboardedAccounts.length > 1) {
    url += '&showAccounts=true';
  }

  if (!_.isEmpty(params)) {
    url += '&' + $.param(params).replace(/\+/g, '%20');
  }

  if (win.state === 'fullscreen') {
    chrome.tabs.create({
      windowId: win.id,
      url,
      index: tab.index + 1, 
    });
  } else {
    const windowRightEdge = win.left + win.width;
    const windowTop = win.top;

    const width = 615;
    const height = 613;

    chrome.windows.create({
      url,
      width,
      height,
      top: Math.max(windowTop, 25) + 70,
      left: windowRightEdge - width - 20,
      type: 'popup',
    });
  }
}

function _getUserEmail(storage) {
  const onboardedAccounts = getOnboardedAccounts(storage);
  return onboardedAccounts.length === 0
    ? null
    : _preferredOnboardedAccount(storage) || onboardedAccounts[0];
}

function _preferredOnboardedAccount(storage) {
  const preferred = storage['account_last_used_in_popup_' + Environment.get()];
  if (!preferred) return null;

  const hasBeenOnboarded = storage['onboarded_' + preferred + '_' + Environment.get()];
  if (!hasBeenOnboarded) return null;

  return preferred;
}

function getOnboardedAccounts(storage) {
  const onboardedAccounts = [];

  Object.keys(storage).some(function(key) {
    const matches = key.match(new RegExp('^onboarded_(.+)_' + Environment.get() + '$'));
    if (matches && storage[key]) onboardedAccounts.push(matches[1]);
  });

  return onboardedAccounts;
}


const FetchProxy = (function() {
  const RESPONSE_PROPERTIES_WHITELIST = [
    'ok',
    'redirected',
    'status',
    'statusText',
    'type',
    'url',
    'headers',
  ];

  const BODY_PROP = '_body';

  async function getBodyAsType(response, type) {
    try {
      const payload = await response[type]();
      return { success: true, payload };
    } catch (err) {
      return { success: false, payload: bundleError(err) };
    }
  }

  const abortControllers = new Map();

  async function fetchWithTimeout(url, options) {
    if (!options.timeout && !options.signalId) {
      return doFetch(url, options);
    }

    if (options.timeout && (typeof options.timeout !== 'number' || options.timeout <= 0)) {
      throw new TypeError('timeout provided was not a positive number');
    }

    const controller = new AbortController();
    let abortTimer;
    if (options.timeout) {
      abortTimer = setTimeout(() => controller.abort(), options.timeout);
    }

    if (options.signalId) {
      abortControllers.set(options.signalId, controller);
    }

    try {
      const init = {
        ...options,
        signal: controller.signal,
      };
      delete init.timeout;
      delete init.signalId;

      return await doFetch(url, init);
    } finally {
      if (abortTimer) clearTimeout(abortTimer);
      if (options.signalId) abortControllers.delete(options.signalId);
    }
  }

  async function doFetch(url, options) {
    if (isBlobURL(url)) {
      const blob = await getBlobFromObjectURL(url);
      return new Response(blob);
    }

    return fetch(url, options);
  }

  function serializeHeaders(headers) {
    if (!headers || !(headers instanceof Headers)) return {};

    const headersObj = {};
    for (const [name, value] of headers) {
      headersObj[name] = value;
    }
    return headersObj;
  }



  async function handleFetch(request, messageSender, sendResponse) {
    try {
      const { url, options, responseFormat = 'raw' } = request;

      const response = await fetchWithTimeout(url, options);

      const shimmedResponse = _.pick(response, RESPONSE_PROPERTIES_WHITELIST);
      shimmedResponse.headers = serializeHeaders(shimmedResponse.headers);

      const responseType = responseFormat === 'objectURL' ? 'blob' : 'text';
      const getBodyResult = await getBodyAsType(response, responseType);

      if (getBodyResult.success && responseFormat === 'objectURL') {
        if (!shimmedResponse.url && isBlobURL(url)) shimmedResponse.url = url;

        const blob = getBodyResult.payload;
        const blobURL = ResponseCache.cacheBlobResponse(shimmedResponse.url, blob);
        shimmedResponse[BODY_PROP] = { success: true, payload: blobURL };
      } else {
        shimmedResponse[BODY_PROP] = getBodyResult;
      }

      sendResponse({ success: true, payload: shimmedResponse });
    } catch (err) {
      sendResponse({ success: false, payload: bundleError(err) });
    }
  }

  function handleAbortFetch(request, messageSender, sendResponse) {
    try {
      const { signalId } = request;
      const controller = abortControllers.get(signalId);

      if (!controller) {
        sendResponse({
          success: false,
          payload: bundleError('Fetch already completed or aborted'),
        });
        return;
      }

      controller.abort();
      sendResponse({ success: true });
    } catch (err) {
      sendResponse({ success: false, payload: bundleError(err) });
    }
  }

  return {
    handleFetch: wrapAsyncMessageHandler(handleFetch, true),
    handleAbortFetch,
  };
})();


function generateMessageId() {
  const chars = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghiklmnopqrstuvwxyz';
  const length = 17; 
  const buffer = new Uint16Array(length);
  window.crypto.getRandomValues(buffer);
  let randomstring = '';
  for (let i = 0; i < buffer.length; i++) {
    randomstring += chars[buffer[i] % chars.length];
  }
  return randomstring;
}



(function() {
  function isValidSender(sender) {
    return !!sender.tab;
  }

  function stripParams(url) {
    try {
      const urlObj = new URL(url);
      return `${urlObj.origin}${urlObj.pathname}`;
    } catch (err) {
      if (err instanceof TypeError) return null;
      throw err;
    }
  }

  function getSenderTags(sender) {
    return {
      senderUrl: sender && stripParams(sender.url),
      tabUrl: sender && sender.tab && stripParams(sender.tab.url),
    };
  }

  function logInvalidMessage(method, sender, request) {
    error.captureMessage(`unexpected message ${method} from content script`, {
      culprit: 'Received unexpected message from content script',
      extra: {
        method,
        sender,
        request,
      },
      tags: getSenderTags(sender),
    });
  }

  function logUnknownMethod(method, sender, request) {
    error.captureMessage(`Unknown method ${method}`, {
      culprit: 'Received unknown request method from content script',
      extra: {
        method,
        sender,
        request,
      },
      tags: getSenderTags(sender),
    });
  }

  const salesforceDomainRegex = /(^|\.)(sales)?force\.com$/;
  const extensionURL = chrome.runtime.getURL('/').slice(0, -1);
  const gmailDomain = 'https://mail.google.com';

  function isSalesforceDomain(hostname) {
    if (!hostname) return false;
    if (salesforceDomainRegex.test(hostname)) return true;

    return false;
  }

  function isWhitelistedUrl(url) {
    const { protocol, origin, hostname } = new URL(url);

    if (origin === extensionURL) return true;

    return (
      /https/.test(protocol) &&
      (origin === gmailDomain || isSalesforceDomain(hostname) || hostname.endsWith('.mixmax.com'))
    );
  }

  const UNIVERSAL_METHODS = new Set([
    'sidebarEnabled',
    'setStorage',
    'contactSelected',
    'sidebarLoaded',
    'closeSidebar',
    'listenerReady',
    'mailTo',
    'callTo',
  ]);

  function isUniversalMethod(method) {
    return UNIVERSAL_METHODS.has(method);
  }





  function handleExtensionMessageBusRequest(request, messageSender) {
    ExtensionMessageBus.trigger(request, messageSender);
    return false;
  }

  const MESSAGE_HANDLERS = new Map([
    ['fetchResource', FetchProxy.handleFetch.bind(FetchProxy)],
    ['abortFetch', FetchProxy.handleAbortFetch.bind(FetchProxy)],
    ['uploadFileToS3', UploaderProxy.handleUploadFileToS3.bind(UploaderProxy)],
    ['sidebarEnabled', handleExtensionMessageBusRequest],
    ['contactSelected', handleExtensionMessageBusRequest],
    ['sidebarLoaded', handleExtensionMessageBusRequest],
    ['closeSidebar', handleExtensionMessageBusRequest],
    ['postNotification', handleExtensionMessageBusRequest],
    ['ping', handleExtensionMessageBusRequest],
    ['refreshGmail', handleExtensionMessageBusRequest],
    ['setStorage', handleExtensionMessageBusRequest],
    ['listenerReady', handleExtensionMessageBusRequest],
    ['mailTo', wrapAsyncMessageHandler(handleMailToRequest, false)],
    ['callTo', wrapAsyncMessageHandler(handleCallToRequest, false)],
  ]);

  function handleMessage(request, messageSender, sendResponse) {
    const { method } = request;
    if (!MESSAGE_HANDLERS.has(method)) {
      logUnknownMethod(method, messageSender, request);
      return false;
    }
    const handler = MESSAGE_HANDLERS.get(method);
    return handler(request, messageSender, sendResponse);
  }

  function listenForMessages() {
    chrome.runtime.onMessage.addListener((request, sender, sendResponse) => {
      try {
        if (!isValidSender(sender)) {
          return false;
        }
        if (!isWhitelistedUrl(sender.tab.url) && !isUniversalMethod(request.method)) {
          logInvalidMessage(request.method, sender, request);
          return false;
        }

        return handleMessage(request, sender, sendResponse);
      } catch (err) {
        error.captureException(err, {
          culprit: 'Failure when handling message in background page',
        });

        return false;
      }
    });
  }

  const CONNECTION_HANDLERS = new Map([
    ['mixmax_cloudinary_upload', UploaderProxy.handlePortConnect.bind(UploaderProxy)],
  ]);

  function isWhiteListedPortName(name) {
    return CONNECTION_HANDLERS.has(name);
  }

  function listenForConnections() {
    chrome.runtime.onConnect.addListener((port) => {
      try {
        const { sender, name: portName } = port;
        if (!isWhitelistedUrl(sender.tab.url) || !isWhiteListedPortName(portName)) {
          logInvalidMessage(portName, sender, null);
          return;
        }

        const handler = CONNECTION_HANDLERS.get(portName);
        handler(port);
      } catch (err) {
        error.captureException(err, {
          culprit: 'Failure when handling connection in background page',
        });
      }
    });
  }

  try {
    listenForMessages();
    listenForConnections();
  } catch (err) {
    error.captureException(err, {
      culprit: 'Failure when attaching chrome.runtime listeners',
    });
  }
})();


function handleMailToRequest(request, messageSender) {
  openMailtoLink(messageSender.tab, request.payload.href);
}

chrome.tabs.onCreated.addListener(function(tab) {
  if (!/^mailto:/.test(tab.url)) return;

  chrome.storage.local.get(null , function(storage) {
    const hasAtLeastOneAccount = getOnboardedAccounts(storage).length > 0;
    if (hasAtLeastOneAccount) {
      openMailtoLink(tab, tab.url);
      chrome.tabs.remove(tab.id);
    }
  });
});

function openMailtoLink(tab, href) {
  let to = href.match(/mailto:([^\?]+)/);
  if (to) to = decodeURIComponent(to[1]);

  const cc = getParameterByName(href, 'cc');
  const bcc = getParameterByName(href, 'bcc');
  const subject = getParameterByName(href, 'subject');
  const body = getParameterByName(href, 'body');

  const payload = {};
  if (to) payload.to = to;
  if (cc) payload.cc = cc;
  if (bcc) payload.bcc = bcc;
  if (subject) payload.subject = subject;
  if (body) payload.body = body;

  chrome.windows.get(tab.windowId, function(win) {
    chrome.storage.local.get(null , function(storage) {
      openComposeWindow(win, tab, storage, payload);
    });
  });
}

function getParameterByName(href, name) {
  const match = RegExp('[?&]' + name + '=([^&]*)', 'i').exec(href);
  return match && decodeURIComponent(match[1].replace(/\+/g, ' '));
}


const LAST_SEEN_NOTIFICATIONS = new Map();
const MAXIMUM_RETAINED_NOTIFICATIONS = 100;

ExtensionMessageBus.on('postNotification', function(notificationOptions, notificationSender) {
  const notificationId = notificationOptions.id;
  if (LAST_SEEN_NOTIFICATIONS.has(notificationId)) return;

  const handlerOptionKeys = ['activateOnClick'];
  const context = {
    clientContext: notificationOptions.context,
    sender: notificationSender,
    handlerOptions: _.map(notificationOptions.buttons, function(button) {
      return _.pick(button, handlerOptionKeys);
    }),
  };
  LAST_SEEN_NOTIFICATIONS.set(notificationId, context);
  if (LAST_SEEN_NOTIFICATIONS.size > MAXIMUM_RETAINED_NOTIFICATIONS) {
    LAST_SEEN_NOTIFICATIONS.delete(LAST_SEEN_NOTIFICATIONS.keys().next().value);
  }

  chrome.notifications.create(
    notificationId,
    {
      type: 'basic',
      iconUrl: 'src/assets/img/icon.png',
      title: notificationOptions.title,
      message: notificationOptions.message,
      buttons: _.map(notificationOptions.buttons, function(button) {
        return _.omit(button, handlerOptionKeys);
      }),
      isClickable: notificationOptions.showClickable,
    },
    function() {
      const err = chrome.runtime.lastError;
      if (err) {
        error.captureException(err, {
          culprit: 'Could not post notification:',
        });
      }
    }
  );
});

function onNotificationClicked(notificationId, buttonIndex) {
  if (buttonIndex === undefined) {
    buttonIndex = null;
  }

  chrome.notifications.clear(notificationId, function() {});

  const context = LAST_SEEN_NOTIFICATIONS.get(notificationId);
  if (!context) {
    return;
  }
  LAST_SEEN_NOTIFICATIONS.delete(notificationId);

  const notificationOptions = {
    activateAfterSending: true,
  };

  if (buttonIndex !== null) {
    const handlerOptions = context.handlerOptions[buttonIndex];
    notificationOptions.activateAfterSending = handlerOptions.activateOnClick;
  }

  ExtensionMessageBus.send(
    'notificationClicked',
    {
      id: notificationId,
      context: context.clientContext,
      buttonIndex,
    },
    context.sender,
    notificationOptions
  );
}
chrome.notifications.onClicked.addListener(onNotificationClicked);
chrome.notifications.onButtonClicked.addListener(onNotificationClicked);

function handleCallToRequest(request, messageSender) {
  openTelLink(messageSender.tab, request.payload.href);
}

function generateCallId() {
  const chars = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghiklmnopqrstuvwxyz';

  const length = 17;
  const buffer = new Uint16Array(length);

  window.crypto.getRandomValues(buffer);

  let randomstring = '';
  for (let i = 0; i < buffer.length; i++) {
    randomstring += chars[buffer[i] % chars.length];
  }
  return randomstring;
}

function openTelLink(tab, href) {
  let phoneNumber = href.match(/tel:([^\;]+)/);
  if (phoneNumber) phoneNumber = decodeURIComponent(phoneNumber[1]);

  chrome.windows.get(tab.windowId, function(win) {
    chrome.storage.local.get(null , function(storage) {
      openCallToWindow(win, tab, storage, { phoneNumber });
    });
  });
}

function openCallToWindow(win, tab, storage, params) {
  const onboardedAccounts = getOnboardedAccounts(storage);
  if (onboardedAccounts.length === 0) {
    return void refreshBrowser('gmail', true );
  }

  const userEmail = _getUserEmail(storage);

  const callId = generateCallId();
  let url = `${Environment.getComposeUrl()}/call/${callId}?user=${encodeURIComponent(userEmail)}`;

  if (!_.isEmpty(params)) {
    url += '&' + $.param(params).replace(/\+/g, '%20');
  }

  if (win.state === 'fullscreen') {
    chrome.tabs.create({
      windowId: win.id,
      url,
      index: tab.index + 1, 
    });
  } else {
    const windowRightEdge = win.left + win.width;
    const windowTop = win.top;

    const width = 615;
    const height = 613;

    chrome.windows.create({
      url,
      width,
      height,
      top: windowTop + 70,
      left: windowRightEdge - width - 20,
      type: 'popup',
    });
  }
}

const extensionOrigin = chrome.runtime.getURL('/').slice(0, -1);

function isNullOriginHeader(name, value) {
  return /^Origin$/i.test(name) && value === 'null';
}

function hasNullOrigin(requestHeaders) {
  for (const { name, value } of requestHeaders) {
    if (isNullOriginHeader(name, value)) return true;
  }
  return false;
}

function shouldOverwriteOriginHeader(requestDetails) {
  const { initiator, tabId, requestHeaders } = requestDetails;
  return initiator === extensionOrigin && tabId === -1 && hasNullOrigin(requestHeaders);
}

function overWriteNullOriginHeader(requestHeaders) {
  return requestHeaders.map(({ name, value }) => {
    return {
      name,
      value: isNullOriginHeader(name, value) ? extensionOrigin : value,
    };
  });
}

chrome.webRequest.onBeforeSendHeaders.addListener(
  (requestDetails) => {
    if (shouldOverwriteOriginHeader(requestDetails)) {
      notifyWorkaroundToStatsd();
      return { requestHeaders: overWriteNullOriginHeader(requestDetails.requestHeaders) };
    }
    return {};
  },
  {
    urls: ['https://' + '*.mixmax.com/' + '*'],
    types: ['xmlhttprequest'],
  },
  ['blocking', 'requestHeaders']
);

let counter = 0;
function notifyWorkaroundToStatsd() {
  if (counter === 0) setTimeout(sendCounterToStatsd, 5000);
  counter++;
}

const chromeVersion = (/Chrome\/([0-9]+)/.exec(navigator.userAgent) || ['', '0'])[1];

function sendCounterToStatsd() {
  if (counter === 0) return; 

  const request = new XMLHttpRequest();
  request.open('POST', Environment.getGatewayUrl() + '/statsd');
  request.setRequestHeader('Content-Type', 'application/json');
  request.send(
    `{"prefix":"extension","increment":[{"name":"overwriteNullOrigin.bugWorkedAround.chrome${chromeVersion}","value":${counter}}]}`
  );
  counter = 0;
}

ExtensionMessageBus.on('ping', function(payload, sender, callback) {
  if (callback) callback('pong');
});


async function findTabsByName(target) {
  if (target !== 'gmail' && target !== 'salesforce' && target !== 'app-home') {
    return Promise.resolve([]);
  }

  const appUrl = new URL(Environment.getAppUrl());

  return new Promise((resolve) => {
    const tabs = [];

    chrome.windows.getAll(
      {
        populate: true,
      },
      (windows) => {
        windows.forEach(function(win) {
          win.tabs.forEach(function(tab) {
            const url = new URL(tab.url);
            const hostname = url.hostname;

            const isSalesforceUrl =
              hostname === 'force.com' ||
              hostname === 'salesforce.com' ||
              hostname.endsWith('.force.com') ||
              hostname.endsWith('.salesforce.com');

            if (target === 'gmail') {
              if (hostname !== 'mail.google.com') return;
            } else if (target === 'salesforce') {
              if (!isSalesforceUrl) return;
            } else if (target === 'app-home') {
              if (hostname !== appUrl.hostname) return;
            } else {
              return; 
            }

            tabs.push(tab);
          });
        });

        resolve(tabs);
      }
    );
  });
}

function refreshBrowser(target, bringToForeground = false) {
  findTabsByName(target).then((matchingTabs) => {
    if (matchingTabs.length > 0) {
      matchingTabs.forEach(({ id }) => chrome.tabs.reload(id));

      if (bringToForeground && target === 'gmail') {
        chrome.tabs.update(matchingTabs[0].id, { active: true });
      }
    } else if (bringToForeground && target === 'gmail') {
      chrome.tabs.create({
        url: 'https://mail.google.com',
        active: true,
      });
    }
  });
}

chrome.runtime.onInstalled.addListener(async function(details) {
  refreshBrowser('salesforce');
  refreshBrowser('gmail');
  refreshBrowser('app-home');

  if (details.reason === 'install') {
    const openAppTabs = await findTabsByName('app-home');
    const openGmailTabs = await findTabsByName('gmail');

    if (openAppTabs.length === 0) {
      const appUrl = new URL(Environment.getAppUrl());
      appUrl.searchParams.set('ref', 'install');
      chrome.tabs.create({
        url: appUrl.toString(),
        active: true,
      });
    } else if (openGmailTabs.length > 0) {
      chrome.tabs.update(openGmailTabs[0].id, { active: true });
    } else {
      chrome.tabs.create({
        url: 'https://mail.google.com',
        active: true,
      });
    }
  }
});

ExtensionMessageBus.on('refreshGmail', function(payload, sender, callback) {
  refreshBrowser('salesforce');
  refreshBrowser('gmail', payload );
  if (callback) callback();
});

ExtensionMessageBus.on('setStorage', function(payload, sender, callback) {
  chrome.storage.local.set(payload);
  ExtensionMessageBus.send('setStorage', payload, {
    notId: sender.id,
  });
  if (callback) callback();
});

function setUninstallUrl() {
  if (!chrome.runtime.setUninstallURL) return;

  chrome.storage.local.get(null , function(storage) {
    let allAccounts = [];
    let loggedInAccounts = [];

    Object.keys(storage).forEach(function(key) {
      let matches = key.match(/^onboarding_started_(.+)_[a-z]+$/);
      if (matches) allAccounts.push(matches[1]);

      matches = key.match(/^has_been_onboarded_(.+)_[a-z]+$/);
      if (matches && !_.contains(allAccounts, matches[1])) allAccounts.push(matches[1]);

      matches = key.match(/^onboarded_(.+)_[a-z]+$/);
      if (matches) loggedInAccounts.push(matches[1]);
    });

    allAccounts = _.unique(allAccounts);
    loggedInAccounts = _.unique(loggedInAccounts).map(function(account) {
      return allAccounts.indexOf(account);
    });

    const url =
      Environment.getAppUrl() +
      '/uninstall-extension' +
      '?a2=' +
      LZString.compressToEncodedURIComponent(allAccounts.join(',')) +
      '&l2=' +
      LZString.compressToEncodedURIComponent(loggedInAccounts.join(','));

    chrome.runtime.setUninstallURL(url);
  });
}

setUninstallUrl();

chrome.storage.onChanged.addListener(setUninstallUrl);
