const BUILD_TARGETS = [];
const SFDC_DOMAIN = /(^|\.)(?:sales)?force\.com$/;
const hostname = window.location.hostname;
const isGmail = hostname === 'mail.google.com';
const isLinkedIn =
  hostname === 'www.linkedin.com' && window.location.pathname.startsWith('/sales/widget');

if (isGmail) {
  BUILD_TARGETS.push('gmail-inbox');
} else if (SFDC_DOMAIN.test(hostname)) {
  BUILD_TARGETS.push('salesforce');
} else if (isLinkedIn) {
  BUILD_TARGETS.push('linkedin');
}

const LOAD_FAILURE_WARN_ORIGINS = ['https://mail.google.com'];

function isValidWarningOrigin(origin) {
  if (!origin.startsWith('https://')) return false;

  if (LOAD_FAILURE_WARN_ORIGINS.includes(origin)) return true;

  return SFDC_DOMAIN.test(origin);
}

const warnEnvironments = new Set(['staging']);

const rValidEnvironment = /^[0-9a-z-]+$/i;

function getExtensionOrigin(environment) {
  if (warnEnvironments.has(environment)) {
    console.warn(`Mixmax extension is running from ${environment}.`);
  }

  if (!environment || environment === 'production') {
    return 'https://extension.mixmax.com';
  }

  if (!rValidEnvironment.test(environment)) {
    throw new Error('extension origin not valid');
  }

  return `https://extension-${environment}.mixmax.com`;
}

const GET_EVERYTHING = null;
chrome.storage.local.get(GET_EVERYTHING, function(storage) {
  const { environment } = storage;
  const extensionOrigin = getExtensionOrigin(environment);

  if (storage.environment === 'local') {
    loadLocal(environment, extensionOrigin);
  } else {
    const sourceSubdir = storage.testSourceSubdir || null;
    loadApp(environment, extensionOrigin, sourceSubdir);
  }
});

async function loadLocal(environment, extensionOrigin) {
  try {
    await loadCSS(environment, extensionOrigin);

    if (isGmail) {
      await pageInteropLoad;
    }

    const sourceSubdir = null;
    await loadScript(environment, extensionOrigin, sourceSubdir);
  } catch (err) {
    if (isValidWarningOrigin(window.location.origin)) {
      alert('Failed to load extension locally. Check Chrome DevTools console for errors.');
    }
    throw err;
  }
}

async function loadApp(environment, extensionOrigin, sourceSubdir) {
  try {
    await loadCSS(environment, extensionOrigin, 6 );
  } catch (err) {
    throw new Error('Failed to load Mixmax CSS');
  }

  if (isGmail && typeof pageInteropLoad !== 'undefined') {
    await pageInteropLoad;
  }

  try {
    await loadScript(environment, extensionOrigin, sourceSubdir, 3 );
  } catch (err) {
    error.captureException(err, {
      culprit: 'Loaded CSS but could not load JS',
      tags: {
        user: getLikelyUserEmail(),
        domain: hostname,
        extensionOrigin,
      },
    });
  }
}

async function retry(asyncFn, maxAttempts = 1, attemptNum = 0) {
  if (attemptNum >= maxAttempts - 1) {
    return asyncFn();
  }

  try {
    return await asyncFn(attemptNum);
  } catch (err) {
    if (err && err.isPermanent) {
      throw err;
    }

    await new Promise((resolve) => setTimeout(resolve, attemptNum * 250));
    return retry(asyncFn, maxAttempts, attemptNum + 1);
  }
}

function loadCSSLink(environment, origin, buildTarget) {
  const cssURL = `${origin}/src/styles-${buildTarget}.css`;

  const styleContainer = document.head || document.documentElement;

  const link = document.createElement('link');
  link.href = cssURL;
  link.type = 'text/css';
  link.rel = 'stylesheet';
  styleContainer.appendChild(link);
}

async function fetchCSS(url) {
  const res = await fetch(url, {
    mode: 'cors',
    credentials: 'omit',
    redirect: 'follow',
    referrerPolicy: 'origin',
  });

  if (!res.ok) {
    throw new Error('failed to fetch CSS');
  }

  return res.text();
}

function loadCSS(environment, origin, numRetries = 0) {
  return Promise.all(
    BUILD_TARGETS.map(async (buildTarget) => {
      if (!numRetries && buildTarget === 'gmail-inbox') {
        return void loadCSSLink(environment, origin, buildTarget);
      }

      const cssURL = `${origin}/src/styles-${buildTarget}.css`;

      const cssRes = await retry(() => fetchCSS(cssURL), numRetries + 1 );

      const styleContainer = document.head || document.documentElement;

      const style = document.createElement('style');
      style.textContent = cssRes;
      styleContainer.appendChild(style);
    })
  );
}

function validateSourceSubdir(sourceSubdir) {
  return (
    typeof sourceSubdir === 'string' &&
    sourceSubdir.startsWith('test-') &&
    /^[0-9A-Z_-]+$/i.test(sourceSubdir)
  );
}

function loadScript(environment, origin, sourceSubdir, numRetries = 0) {
  const maybeSourceSubdir = validateSourceSubdir(sourceSubdir) ? `${sourceSubdir}/` : '';

  return Promise.all(
    BUILD_TARGETS.map(async (buildTarget) => {
      const jsURL = new URL(`${origin}/src/${maybeSourceSubdir}build-${buildTarget}.js`);

      await retry(async (num) => {
        const url = new URL(jsURL);
        if (num) {
          url.searchParams.set('v', `mixmax-extension-${num}`);
        }
        try {
          await import(url.href);
        } catch (err) {
          if (err && err[Symbol.for('mixmaxErrorSource')] === 'mixmax-extension-source') {
            err.isPermanent = true;
          }
          throw err;
        }
      }, numRetries + 1 );
    })
  );
}

function getLikelyUserEmail() {
  const emailRegex = /[a-z0-9\._+\-]+@[a-z0-9\.\-]+\.[a-z]{2,20}/i;
  const messageElement = document.querySelector('.msg');
  if (messageElement && emailRegex.test(messageElement.textContent)) {
    return messageElement.textContent.match(emailRegex)[0];
  }
  const titleMatch = document.title.match(emailRegex);
  return titleMatch ? titleMatch[0] : undefined;
}
